from typing import Any, Generator

def capture_kwargs_by_type(
    kwargs: dict, types: tuple[type, ...] | type
) -> Generator[tuple[Any, Any], None, None]:
    """
    Loop through a dictionary's items that match type(s)
    """
    for key, val in kwargs.items():
        if isinstance(val, types):
            yield (key, val)


def capture_args_by_type(
    args: list | tuple, types: tuple[type, ...] | type
) -> Generator[tuple[int, Any], None, None]:
    """
    Enumerate through an iterable's values that match type(s)
    """
    for i, arg in enumerate(args):
        if isinstance(arg, types):
            yield (i, arg)


def combine_args_and_children_to_list(args: tuple, children: list | None) -> list:
    """
    Combines elements passed as positionals with elements passed
    inside `children` into one list.
    """
    args = list(args)
    if children is not None:
        args += children
    return args


def move_dict_args_to_other_dict(container: list, other_dict: dict) -> None:
    """
    If any dictionaries are in `container` (likely because they were passed
    as positional arguments), they will be popped from container and their
    key/val pairs safely set to other_dict.

    Mutates inplace:
        `container`
        `other_dict`
    """
    for i, elem in enumerate(container):
        if isinstance(elem, dict):
            new = container.pop(i)
            for key, val in new.items():
                if other_dict.get(key) is None:
                    other_dict[key] = val


def convert_all_to_type(
    container: list,
    from_type: type | tuple[type, ...],
    to_type: type,
    strict: bool = False,
) -> None:
    def check_elem(elem) -> bool:
        if strict is True:
            if isinstance(from_type, tuple):
                return type(elem) in from_type
            return type(elem) == from_type
        return isinstance(elem, from_type)

    for i, elem in enumerate(container):
        if check_elem(elem):
            container[i] = to_type(elem)


def convert_sibling_types(instance, container: list) -> None:
    elem_type = type(instance).elem_type
    sib_type = elem_type.sibling_type
    for i, elem in enumerate(container):
        if isinstance(elem, sib_type):
            container[i] = elem.transpose()


def move_remaining_kwargs_to_dict(kwargs: dict, to_dict: dict, safely: bool = False) -> None:
    """
    OVERRIDES values in to_dict
    """
    for key, val in kwargs.items():
        if safely is True and key in to_dict:
            continue
        to_dict[key] = val



from __future__ import annotations

from pandas import DataFrame
from collections import namedtuple, ChainMap
from copy import copy
from excelbird.exceptions import SchemaError
from typing import overload


# namedtuple is like a tuple, but you can access elements by name. So
# `Column` is the data structure to be held as dict values by a `Schema`
Column = namedtuple("Column", "input, output")


class Schema(dict):
    """
    A subclass of dictionary, designed to hold the schema for imported data.
        Keys: python-friendly variable names
        Values: tuple: ("Input col name", "Output col name")

    Its purpose is to decouple column naming from the user's workflow.
    For each variable we want to use in our code, we'll store the following:

    "Input Col Name"   ->   var_name   ->   "Output Col Name"
                               ^
                        this never changes

    If we define this info ahead of time, our program only ever needs to reference
    the `var_name`. Input data and output format can change freely without breaking our code.

    This class has all the tools needed to convert between naming formats.
    Use `select_inputs()` and/or `read_excel()` when reading input data to `var_name`s.
    `vars_to_outputs()` returns a dict you can use to refactor columns to their output format.
    """

    def __init__(
        self, *schemas, **kwargs: tuple[str, str] | tuple[str] | list[str] | str
    ) -> None:
        """Accepts keyword args only to enforce that var names are valid python"""
        if not all(isinstance(s, Schema) for s in schemas):
            raise TypeError("Positional args can only be existing Schemas")
        # Convert all passed values to tuples
        kwargs = {
            k: v if isinstance(v, tuple) else tuple(v) if isinstance(v, list) else (v,)
            for k, v in kwargs.items()
        }
        # For values of length 1 (input column only), infer the output column by duplicating it
        kwargs = {k: v if len(v) > 1 else v + v for k, v in kwargs.items()}

        if any(len(val) > 2 for val in kwargs.values()):
            raise ValueError(
                "Values must be a max of length 2. One input col name, and one output col name"
            )
        # Convert to Column
        kwargs = {k: Column(*v) for k, v in kwargs.items()}
        # If other schemas were passed in, create those as well
        # ChainMap combines dictionaries. We reverse the input first,
        # since for some reason ChainMap returns the values in opposite order
        super().__init__(**ChainMap(*tuple(reversed(schemas))), **kwargs)

    def __getattr__(self, key: str) -> Column:
        """Lets you access dict items with dot notation"""
        if key in self.keys():
            return self[key]
        raise KeyError(f"Unknown key, '{key}'")

    @overload
    def __getitem__(self, key: list) -> Schema:
        ...

    @overload
    def __getitem__(self, key: str) -> Column:
        ...

    def __getitem__(self, key) -> Column | Schema:
        """
        Acts normal, unless you past a list.
        If a list is passed, filter and re-order the schema just like a
        dataframe. Returns new object
        """
        if not isinstance(key, list):
            return super().__getitem__(key)

        missing = [k for k in key if k not in self.keys()]
        if len(missing) > 0:
            raise SchemaError(f"Keys {missing} not present in schema")

        reordered = {copy(k): copy(self[k]) for k in key}
        return type(self)(**reordered)

    def __setitem__(self, key: str, val: Column | tuple[str, ...] | str) -> None:

        if isinstance(val, Column):
            return super().__setitem__(key, val)

        if isinstance(val, (list, tuple)):
            new = Column(*val)
            return super().__setitem__(key, new)

        if isinstance(val, str):
            new = Column(val, val)
            return super().__setitem__(key, new)

        raise ValueError(f"Invalid value, {val}")

    def drop(self, columns: list[str] | str) -> Schema:
        """
        Drop the specified keys
        """
        if not isinstance(columns, (list, tuple)):
            columns = [columns]

        return type(self)(
            **{copy(k): copy(v) for k, v in self.items() if k not in columns}
        )

    def apply(self, df: DataFrame, strict: bool = False) -> DataFrame:
        """
        - Filter ``df`` to remove columns that aren't in self
        - Re-order columns according to self's order

        If ``strict=True``, raise an error if ``df`` doesn't at least contain
        all columns specified in self.
        """
        if strict is False:
            return df[[k for k in self.keys() if k in df.columns]].copy()
        try:
            return df[[k for k in self.keys()]].copy()
        except KeyError:
            missing = [k for k in self.keys() if k not in df.columns]
            raise KeyError(
                f"Schema apply strict: The following columns were not found "
                f"in the dataframe (did you forget to run .select_inputs() first?): {missing}"
            )

    def rename(
        self,
        keys: dict | None = None,
        inputs: dict | None = None,
        outputs: dict | None = None,
    ) -> Schema:
        """
        Rename any part of self's items (key, input, output), by passing a dict
        who's keys are a current key in self, and values are the updated element.
        """
        new = self.copy()
        if keys is not None:
            new = type(self)()
            for key, val in self.items():
                if key not in keys:
                    new[key] = val
                else:
                    new[keys[key]] = val

        if inputs is not None:
            for key, new_name in inputs.items():
                new[key] = Column(new_name, new[key].output)

        if outputs is not None:
            for key, new_name in outputs.items():
                new[key] = Column(new[key].input, new_name)

        return new

    def update(self, new: dict | None = None, **kwargs) -> None:
        """
        Like ``dict.update()``, but if a regular ``dict`` or kwargs are passed,
        they're first used to create a new ``Schema`` before updating, so the
        correct format is maintained
        """
        if isinstance(new, type(self)):
            return super().update(new)
        if new is not None:
            return super().update(type(self)(**new))
        return super().update(type(self)(**kwargs))

    def rename_inputs_to_vars(self, df: DataFrame) -> DataFrame:
        """
        Take a dataframe in its input format and rename its columns to the
        python-friendly keys in self.
        """
        return df.rename(columns={val.input: key for key, val in self.items()})

    def rename_vars_to_outputs(self, df: DataFrame) -> DataFrame:
        """
        Rename columns to self's output names
        """
        return df.rename(columns={key: val.output for key, val in self.items()})

    def inputs(self) -> list[str]:
        return [val.input for val in self.values()]

    def outputs(self) -> list[str]:
        return [val.output for val in self.values()]

    def select_inputs(self, df: DataFrame) -> DataFrame:
        """
        Renames desired columns to var names, and selects them.
        If a column isn't found, an error is raised to force you to correct
        your schema.
        """
        missing = [col for col in self.inputs() if col not in df.columns]
        if len(missing) > 0:
            raise SchemaError(
                f"Schema requires input column(s), {missing}, not found in data."
            )
        df = self.rename_inputs_to_vars(df)
        return df[[k for k in self.keys()]]

    def select_outputs(self, df: DataFrame) -> DataFrame:
        """
        Renames var columns to their output names, and selects them.
        If any columns are missing, an error is raised to remind you to create them.
        """
        missing = [col for col in self.keys() if col not in df.columns]
        if len(missing) > 0:
            raise SchemaError(f"Please add columns, {missing} before outputting.")
        df = self.rename_vars_to_outputs(df)
        return df[[k for k in self.outputs()]]

    def reset_inputs(self) -> Schema:
        """
        Sets all inputs with output values. Use this if you're
        using a previous schema to read in data that was outputted from it
        """
        new = self.copy()
        for key in new.keys():
            new[key] = Column(new[key].output, new[key].output)
        return new

    def reset_outputs(self) -> Schema:
        """
        Fills all outputs with the current inputs
        """
        new = self.copy()
        for key in new.keys():
            new[key] = Column(new[key].input, new[key].input)
        return new

    def copy(self) -> Schema:
        return type(self)(**{copy(k): copy(v) for k, v in self.items()})

    def _repr_html_(self):

        return DataFrame(
            list(zip(self.inputs(), self.outputs())),
            columns=["Input", "Output"],
            index=list(self.keys()),
        )._repr_html_()



from excelbird._utils.util import *
from excelbird._utils.argument_parsing import *
from excelbird._utils.cell_util import *
from excelbird._utils.color_algorithms import *
from excelbird._utils.pass_attributes import *
from excelbird._utils.validation import *



"""
Alternate version of fx. The goal of these functions
is simply to make writing complex nested Funcs easier,
so that you don't have to put your entire Func string inside
one long string.

Instead of...

'''
MAX(
  NETWORKDAYS(
    MAX({forecast_start_date}, {date_start}),
    {date_end},
    IF(
      OR({loc}="Offshore", {loc}="Nearshore"),
      0,
      {[holidays].range()}
    )
  ),
  0
)
'''

You can now use:

MAX(
    NETWORKDAYS(
        MAX("{forecast_start_date}, {date_start}"),
        "{date_end}",
        IF(
            OR("{loc}='Offshore', {loc}='Nearshore'"),
            0,
            "{[holidays].range()}"
        )
    ),
    0
)

"""



from excelbird.core.cell import *
from excelbird.core.series import *
from excelbird.core.frame import *
from excelbird.core.stack import *
from excelbird.core.sheet import *
from excelbird.core.book import *
from excelbird.core.expression import *
from excelbird.core.function import *
from excelbird.core.gap import *
from excelbird.core.item import *
from excelbird.core.merged import *

"""
TODO:
- allow a docstring in functions with triple backtick
- Func needs to have id and header
- DUDE!!!! Make Func use a unique syntax to declare an Expr inside.
  The enclosure, @[], represents an Expr. So @[<stuff>] will be interpreted
  EXACTLY as Expr("<stuff>"). AND since Expr can take a single reference without
  brackets, this gives us the ULTIMATE syntax combo. Inside a Func, reference an element
  like `@[some_elem]`. OR reference an Expr like `@[[foo]+[bar].range()]`
  the bracket references must be prefixed with an @[]
  - A reference is made like @[some_elem]. To insert an Expr, 
- Fix .ref() so that it can handle child elements who are item, expr, func
- ALL elements should be subclass of XBElement, to make type checking easier.
  ALL dynamic elements (gap, func, expr, item) should be subclass of Dynamic
- Ids and headers declared in Expr, Func, and Item should all be placed into global
  memory right away. This way, when Book's .write() is called, we can immediately determine
  if there are any invalid references, without trying to go through the whole process.
- Refactor `astype` to `to_sibling`
- Doc examples and doctests!!
- Store styles in dict instead of Style
- In range references from separate sheets, remove redundant repeated sheet name
- Change expression to @[] instead of [].
- Allow Func to be a single string, parsing out all @[] just like we do with Expr
- Conditional formatting!
- Offer syntax in cell expression to specify which $s to use
- Somehow, figure out how to use column name references.
- Fix table formatting with cross-sheet references
- Make diagrams for documentation

Make cell references smarter, using $ signs where appropriate.
    First change cell expression tree to hold 5 elements instead of 3.
    Second and last element hold a tuple designating how to lock the cell reference.
    (False,False) -> A5, (True,False) -> $A5, (False,True) -> A$5, (True,True) -> $A$5
    Then, in the math module, determine what these locks should be.
        - Create an express_with_lock function, because we can't express
          locks in a "a + b" expression (3 elements only), so instead, intermediate
          math will be done through a function that can take locks. 
        - LOGIC:
            - The lock is only set once. So if elem_math recieves a lock for an element,
              keep it. 
            - If referencing a Cell -> (True, True)
            - Col -> (True, False)
            - Row -> (False, True)
    For Func, we need a way to choose to specify lock. I think we can do this by allowing
    a 'lock' param inside .range(). Make sure Cell has a .range() as well now. This will
    be a boolean to full-lock or not. It should default to True, for all versions of .range()
"""

